// -*- c++ -*-
//------------------------------------------------------------------------------
//                            PidFileLock.cpp
//------------------------------------------------------------------------------
//  Copyright (c) 2001 by Vladislav Grinchenko
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Library General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
//------------------------------------------------------------------------------

//System Includes
#include <errno.h>		// errno
#include <string.h>		// strerror(3)
#include <unistd.h>
#include <fcntl.h>
#include <sstream>		
#include <stdio.h>

//Local Includes
#include "assa/CommonUtils.h"
#include "assa/PidFileLock.h"

using namespace ASSA;

//------------------------------------------------------------------------------
// External Events
//------------------------------------------------------------------------------

PidFileLock::
PidFileLock () : 
	m_fd (-1),
	m_error (0),
	m_error_msg ("no errors")
{
	trace_with_mask ("PidFileLock::PidFileLock", PIDFLOCK);

	l_whence = SEEK_SET;
	l_start = l_len = l_pid = 0;
}

PidFileLock::
~PidFileLock ()
{
	trace_with_mask ("PidFileLock::~PidFileLock", PIDFLOCK);

	if (m_fd != -1) {
		unlink (m_filename.c_str ());
		DL((PIDFLOCK,"PID lock file removed.\n"));

		close (m_fd);
		DL((PIDFLOCK,"PID lock file closed.\n"));
	}
}

bool
PidFileLock::
lock (const string& fname_)
{
	trace_with_mask ("PidFileLock::lock", PIDFLOCK);

	m_filename = Utils::strenv (fname_.c_str ());

	DL((PIDFLOCK,"PID lock file: \"%s\"\n", m_filename.c_str ()));

	std::ostringstream mypid;
	int val;
	int len;
	val = len = 0;

	if ((m_fd = ::open (m_filename.c_str (), O_WRONLY | O_CREAT, 0644)) < 0)
	{
		log_error("open() error.");
		return false;
	}
	DL((PIDFLOCK,"PID lock file opened.\n"));

	/*--- See if file is already locked ---*/

	pid_t owner_pid;
	
	if ((owner_pid = test_region ()) == 0) {
		DL((PIDFLOCK,"PID file is not locked\n"));
	}
	else {
		DL((PIDFLOCK,"PID file locked by (%d)\n", owner_pid));
	}

	/*--- Try and set a write lock on the entire file ---*/

	if (lock_region () < 0)	{

		if (errno == EACCES || errno == EAGAIN)	{
			log_error("PID file is locked by another process");
		}
		else {
			log_error("write lock error");
		}
		goto done;
	}
	DL((PIDFLOCK,"PID lock file locked.\n"));

	/*--- truncate to zero length, now that we have the lock ---*/

	if (ftruncate (m_fd, 0) < 0) {
		log_error("ftruncate() error");
		goto done;
	}
	DL((PIDFLOCK,"PID lock file truncated.\n"));

	mypid << getpid () << std::ends;
	len = strlen (mypid.str ().c_str ());
	
	if (write (m_fd, mypid.str ().c_str (), len) != len) {
		log_error("write() error");
		goto done;
	}
	DL((PIDFLOCK,"PID (%d) written to lock file.\n", getpid ()));

	/*-- set close-on-exec flag for descriptor ---*/

	if ((val = ::fcntl(m_fd, F_GETFD, 0)) < 0) {
		log_error("fcntl(F_GETFD) error");
		goto done;
	}
	val |= FD_CLOEXEC;
    
	if (::fcntl (m_fd, F_SETFD, val) < 0) {
		log_error("fcntl(F_SETFD) error");
		goto done;
	}
	DL((PIDFLOCK,"CLOSE-ON-EXEC is set on FD.\n"));

 done:
	if (get_error () != 0) {
		::close (m_fd);
		m_fd = -1;
	}
	return m_error == 0 ? true : false;
}

//------------------------------------------------------------------------------
//   Utility functions
//------------------------------------------------------------------------------

int
PidFileLock::
lock_region ()
{
	trace_with_mask ("PidFileLock::lock_region", PIDFLOCK);
	int ret;

	l_type   = F_WRLCK;
	l_start  = 0;
	l_whence = SEEK_SET;
	l_len    = 0;

	ret = ::fcntl (m_fd, F_SETLK, static_cast<struct flock*>(this));

	DL((PIDFLOCK,"fcntl(fd=%d) returned: %d\n", m_fd, ret));
	return (ret);
}

pid_t
PidFileLock::
test_region ()
{
	trace_with_mask ("PidFileLock::test_region", PIDFLOCK);

	int ret;

	l_type   = F_WRLCK;
	l_start  = 0;
	l_whence = SEEK_SET;
	l_len    = 0;
	
	ret = ::fcntl (m_fd, F_GETLK, static_cast<struct flock*>(this));

	DL((PIDFLOCK,"fcntl(fd=%d) returned: %d\n", m_fd, ret));

	if (ret < 0) {
		log_error("fcntl() error.");
		return 0;
	}

	if (l_type == F_UNLCK) 
	{
		DL((PIDFLOCK,"Region is not locked\n"));
		return(0);	
	}
	DL((PIDFLOCK,"Region is locked by PID %d\n", l_pid));

	return (l_pid);		/* true, return pid of lock owner */
}

void
PidFileLock::
dump (void) 
{
	trace_with_mask("PidFileLock::dump", PIDFLOCK);

	DL((PIDFLOCK,"m_filename : \"%s\"\n", m_filename.c_str()));
	DL((PIDFLOCK,"m_error    : %d\n",     get_error ()));
	DL((PIDFLOCK,"m_error_msg: \"%s\"\n", get_error_msg ()));
	DL((PIDFLOCK,"m_fd       : %d\n",     m_fd));

	if (m_fd == -1) return;

	test_region ();

	if (l_type == F_RDLCK)
		DL((PIDFLOCK,"l_type    : F_RDLCK"));

	if (l_type == F_WRLCK)
		DL((PIDFLOCK,"l_type    : F_WRLCK"));

	if (l_type == F_UNLCK)
		DL((PIDFLOCK,"l_type    : F_UNLCK"));

	DL((PIDFLOCK,"l_whence  : %s\n",
	    l_whence == SEEK_SET ? "SEEK_SET" :
	    l_whence == SEEK_CUR ? "SEEK_CUR" : "SEEK_END"));

	DL((PIDFLOCK,"l_start   : %d\n",   l_start));
	DL((PIDFLOCK,"l_len     : %d\n",   l_len  ));
	DL((PIDFLOCK,"l_pid     : %ld\n",  l_pid  ));
}

void
PidFileLock::
log_error (const char* msg_)
{
	m_error = errno;
	EL((ERROR, "Error: \"Failed to get a lock on PID file\"\n"));
}


